"""Utilities for gRPC services."""
import logging
from functools import wraps
from typing import Any, Callable, Iterator, TypeVar

from google.protobuf.message import Message
from grpc import ServicerContext

from .exceptions import ServiceError
from .utils import format_message, get_call_id

RequestType = TypeVar('RequestType', Message, Iterator[Message])


def wrap_unary(func: Callable[[Any, RequestType, ServicerContext], Message]
               ) -> Callable[[Any, RequestType, ServicerContext], Message]:
    """Decorate a gRPC service method which returns a single message."""
    @wraps(func)
    def wrapper(self: Any, request: RequestType, context: ServicerContext) -> Message:
        call_id = get_call_id()
        logger = logging.getLogger(func.__module__)
        logger.debug("[%s] %s(%s)", call_id, func.__qualname__, format_message(request))
        try:
            reply = func(self, request, context)
        except ServiceError as error:
            logger.info("[%s] %s failed with %r", call_id, func.__qualname__, error)
            context.abort(error.status, error.message)
        except Exception as error:
            logger.error("[%s] %s raised error %r", call_id, func.__qualname__, error)
            raise
        logger.debug("[%s] %s returned %s", call_id, func.__qualname__, format_message(reply))
        return reply
    return wrapper


def wrap_stream(func: Callable[[Any, RequestType, ServicerContext], Iterator[Message]]
                ) -> Callable[[Any, RequestType, ServicerContext], Iterator[Message]]:
    """Decorate a gRPC service method which returns a stream."""
    @wraps(func)
    def wrapper(self: Any, request: RequestType, context: ServicerContext) -> Iterator[Message]:
        call_id = get_call_id()
        logger = logging.getLogger(func.__module__)
        logger.debug("[%s] %s(%s)", call_id, func.__qualname__, format_message(request))
        try:
            reply = func(self, request, context)
            for item in reply:
                logger.debug("[%s] %s returned %s", call_id, func.__qualname__, format_message(item))
                yield item
        except ServiceError as error:
            logger.info("[%s] %s failed with %r", call_id, func.__qualname__, error)
            context.abort(error.status, error.message)
        except Exception as error:
            logger.error("[%s] %s raised error %r", call_id, func.__qualname__, error)
            raise
    return wrapper
